import sys

def main():
    print(sys.executable)
    print(sys.argv)
    args = ['-m', 'millconnect.gui']
    if len(sys.argv) > 1:
        args += sys.argv[1:]
    print(f'respawning: {sys.executable} {args}')
    print(sys.exec_prefix)

if __name__ == "__main__":
    main()